#    .SYNOPSIS
#    Script to remove all Reviewer instances which do not have any user instances attached via the Is User relationship.
#
#    .DESCRIPTION
#    This script is designed to search SCSM for any instances of the Reviewer class which do not have any users assigned via the Is User relationship, then remove said Reviewers.
#
#    Service Manager will "by design" automatically create an unassigned Reviewer instance to any Review activities which, upon creation, lack any assigned reviewers. For example, if a service request is created via a template which contains a review activity which does not have any users specified as reviewers in the template, then upon creation, Service Manager will add a blank Reviewer to the review activity.
#
#    This "feature" was added in a prior version to address issues wherein change requests would seemingly "skip" review activities, as the default behavior was to auto-approve any review activities which lacked reviewers. Rather than changing the default behavior, a workflow was added to create empty Reviewer instances.
#
#    This becomes a problem when leveraging workflows to add reviewers post-request creation, whether those workflows are Orchestrator runbooks or PowerShell scripts.
#
#    Since no reviewer is present upon review activity creation, a blank one is added before the workflow adds the correct reviewer. Once that reviewer approves or denies the request, it languishes, as the unassigned reviewer instance has not voted and thus the approval criteria has not been met.
#
#    This script is intended to run via a workflow (Orchestrator, SMA, an MP, or a Scheduled Task) and automatically search for the blank reviewer instances and remove them, allowing the review activity to be approved or rejected and the overall request to proceed.
#
#    This script REQUIRES the community PowerShell module SMLets to be installed in order to function!
#
#    .NOTES
#    Written by Gabriel Taylor, 2014

#region Variables
$SMServer = $env:COMPUTERNAME
$SMLetsModule = "SMLets"
$ClassName = "System.Reviewer"
$ProjName = "System.ReviewerProjection"
$CritObjType = "Microsoft.EnterpriseManagement.Common.ObjectProjectionCriteria"
#endregion

#region XML Criteria
## Declare the criteria defining the Reviewers without Is User users
$CriteriaHS = @'
<Criteria xmlns="http://Microsoft.EnterpriseManagement.Core.Criteria/">
<Reference Id="System.WorkItem.Activity.Library" PublicKeyToken="31bf3856ad364e35" Version="7.5.0.0" Alias="CoreActivity" />
<Reference Id="System.Library" PublicKeyToken="31bf3856ad364e35" Version="7.5.0.0" Alias="CoreSystem" />
<Expression>
<UnaryExpression>
<ValueExpression>
<GenericProperty Path="$Context/Path[Relationship='CoreActivity!System.ReviewerIsUser' TypeConstraint='CoreSystem!System.User']$">Id</GenericProperty>
</ValueExpression>
<Operator>IsNull</Operator>
</UnaryExpression>
</Expression>
</Criteria>
'@
#endregion

#region Module Import and Preparation
## Set the identity of the default management server for SMLets cmdlets to target to the specified SMServer
$smdefaultcomputer = $SMServer

## Import SMLets PowerShell Module
Write-Verbose "[$(Get-Date)] Loading SMLets PowerShell module ..."
$ExistingModules = (Get-Module | ForEach-Object{$_.name}) -join " "
If(!$ExistingModules.Contains($SMLetsModule)) {
    Import-Module $SMLetsModule -Force -Verbose:$false
}

## Get the Class
Write-Verbose "[$(Get-Date)] Retrieving the $($ClassName) class ..."
$ClassObj = Get-SCClass -ComputerName $SMServer -Name $ClassName 

## Get the type projection
Write-Verbose "[$(Get-Date)] Retrieving the $($ProjName) type projection ..."
$ProjObj = Get-SCSMTypeProjection $ProjName -ComputerName $SMServer

## Create the criteria object
Write-Verbose "[$(Get-Date)] Creating the ObjectProjectionCriteria object ..."
$CritObj = New-Object $CritObjType $CriteriaHS,$ProjObj.__Base,$ProjObj.ManagementGroup
#endregion

#region Collect Instances
## Create an array of the returned instances (here they are PSCustomObjects)
Write-Verbose "[$(Get-Date)] Retrieving any unassigned Reviewer instances ..."
$ProjObjInstances = Get-SCSMObjectProjection -ComputerName $SMServer -criteria $CritObj 

## Check to ensure the array isn't empty, then add the ProjObjInstances' IDs to an array
$IdArray = @()
If($ProjObjInstances.count -ne 0) {
    ForEach($ObjInst in $ProjObjInstances) {
        $ObjId = $ObjInst.ReviewerId
        $IdArray += $ObjId
    }
}
#endregion

#region Remove Bad Instances
## Check to make sure IdArray isn't empty, then get the reviewers as Enterprise Management Instances and remove them 
If($IdArray.count -ne 0) {
    Write-Verbose "[$(Get-Date)] Unassigned Reviewer instances found; removing ..."
    ForEach($Id in $IdArray) {
        Get-SCClassInstance -Class $ClassObj -ComputerName $SMServer -Filter "ReviewerId -eq $Id" | Remove-SCClassInstance
    }
}
Else {
    Write-Verbose "[$(Get-Date)] No Unassigned Reviewer instances found; ending script ..."
}
#endregion
